/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _ALC_H
#define _ALC_H

#include <set>

#include "first.h"
#include "node.h"
#include "graph.h"
#include "inout.h"

extern string* SYM_TEST;

//------------------------------------------------------------------------------

class ALCNode: public Node {
  public:
    ALCNode (Formula* f) { content = f; };

    ~ALCNode() { if (!otherInfo) delete otherInfo; };

    bool chooseRule(labels_queue &labels);
    Formula* applyRule(Formula* label, Formula* generalAssumptions = NULL);
    Formula* computeUnsatCore(Formula* generalAssumptions, Node* child = NULL);
};

//------------------------------------------------------------------------------

class ALCGraph: public Graph {
  protected:
    Node* newNode(Formula* f) { return new ALCNode(f); };
};

//------------------------------------------------------------------------------

class ALCInOut: public InOut {
  private:
    map<string*,Formula*> conceptDefs;
    string* lastConceptName;
  public:
    ALCInOut() { lastConceptName = NULL; };

    void omitSeperators(FILE *pf);
    string* getSym(FILE *pf);
    string* getLastConceptName() { return lastConceptName; };
    int getNumOfConceptDefs() { return conceptDefs.size(); };
    Formula* getConceptDef(string conceptName);

    void init();

    Formula* getFormula(FILE *pf);
    void readAcyclicTBox(FILE *pf);
    void clearConceptDefs() { conceptDefs.clear(); lastConceptName = NULL; };
    Formula* getTESTformula() {
	ASSERT(conceptDefs.find(SYM_TEST) != conceptDefs.end());
	return conceptDefs[SYM_TEST]; 
    };

    void display(Formula* f, FILE *pf);
    void display(ModalIndex* idx, FILE *pf) { idx->display(pf); };
};

//------------------------------------------------------------------------------

class ALC {
  public:
    static float computeWeight(Formula* f);
    static void init();
    static void initChild(int seconds);

    static char checkSatisfiability(Formula* checkedFormula, Formula* generalAssumptions, 
                                    int seconds, bool clearParentFCat = false);

    static void test_batch_file(int command, const char *fileName, int seconds);

    static void main();
};

#endif
