/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include "ALC.h"

//string* SYM_EOLN;
string* SYM_EOF;
string* SYM_PAR_LEFT;
string* SYM_PAR_RIGHT;
string* SYM_ERROR;

string* SYM_AND;
string* SYM_OR;
string* SYM_NOT;
string* SYM_ALL;
string* SYM_SOME;
string* SYM_BOT;
string* SYM_TOP;

string* SYM_DEFINE_PRIMITIVE_ROLE;
string* SYM_DEFINE_PRIMITIVE_CONCEPT;
string* SYM_DEFINE_CONCEPT;
string* SYM_TEST;

#define define_sym(x,y) { s = string(y); x = SymbolCatalogue::find(s); }

void ALCInOut::init()
{
  string s;
//define_sym(SYM_EOLN, "EOLN");
  define_sym(SYM_EOF, "EOF");
  define_sym(SYM_PAR_LEFT, "(");
  define_sym(SYM_PAR_RIGHT, ")");
  define_sym(SYM_ERROR, "ERROR");
  define_sym(SYM_AND, "AND");
  define_sym(SYM_OR, "OR");
  define_sym(SYM_NOT, "NOT");
  define_sym(SYM_ALL, "ALL");
  define_sym(SYM_SOME, "SOME");
  define_sym(SYM_BOT, "*BOTTOM*");
  define_sym(SYM_TOP, "*TOP*");

  define_sym(SYM_DEFINE_PRIMITIVE_ROLE, "DEFINE-PRIMITIVE-ROLE");
  define_sym(SYM_DEFINE_PRIMITIVE_CONCEPT, "DEFINE-PRIMITIVE-CONCEPT");
  define_sym(SYM_DEFINE_CONCEPT, "DEFINE-CONCEPT");
  define_sym(SYM_TEST, "TEST");
};

void ALCInOut::omitSeperators(FILE *pf)
{
  while (!feof(pf)) {
    char ch = fgetc(pf);
    if (ch != ' ' && ch != '\t' && ch != '\n' && ch != ';') {
      ungetc(ch, pf);
      return;
    }
    else if (ch == ';')
      while (!feof(pf) && ch != '\n')
        ch = fgetc(pf);
  };
};

string* ALCInOut::getSym(FILE *pf)
{
  string s;

  omitSeperators(pf);
  char ch = fgetc(pf);

  switch (ch) {
    case EOF : return SYM_EOF;
    case '(' : return SYM_PAR_LEFT;
    case ')' : return SYM_PAR_RIGHT;
  };

  if (!isalnum(ch) && ch != '*') { // '*' is for handling *TOP* and *BOTTOM*
    return SYM_ERROR;
  };
  s.append(1,ch);

  while (!feof(pf)) {
    ch = fgetc(pf);
    if (!isalnum(ch) && ch != '*' && ch != '-') {
      ungetc(ch,pf);
      return SymbolCatalogue::find(s);
    };
    s.append(1,ch);
  };

  return SymbolCatalogue::find(s);
};

Formula* ALCInOut::getFormula(FILE *pf)
{
  string* s = getSym(pf);

  if (s == SYM_PAR_LEFT) {
    s = getSym(pf);

    if (s == SYM_AND || s == SYM_OR) {
      Formula* g1 = getFormula(pf);
      Formula* g2 = getFormula(pf);
      FormulaSet  g3(s == SYM_AND ? AND_SET : OR_SET, g1, g2);
      s = getSym(pf);
      ASSERT(s == SYM_PAR_RIGHT);
      return g3.normalize();
    };

    if (s == SYM_NOT) {
      Formula* g = getFormula(pf);
      s = getSym(pf);
      ASSERT(s == SYM_PAR_RIGHT);
      return g->getNegation();
    };

    if (s == SYM_ALL || s == SYM_SOME) {
      string* s2 = getSym(pf);
      ASSERT(s2->c_str()[0] == 'R');
      ModalIndexSym* midx = ModalIndexSymCat::find(s2);
      Formula* g = getFormula(pf);
      FormulaModalQuant  g2(s == SYM_ALL ? FORALL : EXISTS, midx, g);
      s = getSym(pf);
      ASSERT(s == SYM_PAR_RIGHT);
      return g2.normalize();
    };

    ASSERT(false);
  };

  if (s == SYM_TOP) return FormulaCatalogue::top;
  if (s == SYM_BOT) return FormulaCatalogue::bot;

  if (conceptDefs.find(s) != conceptDefs.end())
    return conceptDefs[s];
  else
    return FormulaCatalogue::find(new FormulaLiteral(POS_LITERAL,s));
};

void ALCInOut::readAcyclicTBox(FILE *pf)
{
  string* s = getSym(pf);
  ASSERT(s == SYM_PAR_LEFT);
  s = getSym(pf);

  while (s != SYM_PAR_RIGHT) {
    ASSERT(s == SYM_PAR_LEFT);
    s = getSym(pf);

    if (s == SYM_DEFINE_PRIMITIVE_ROLE || s == SYM_DEFINE_PRIMITIVE_CONCEPT) {
      s = getSym(pf); // read and ignore the role/concept name
      s = getSym(pf);
      ASSERT(s == SYM_PAR_RIGHT);
      s = getSym(pf);
      continue;
    };

    if (s == SYM_DEFINE_CONCEPT) {
      Formula* lhs = getFormula(pf);
      Formula* rhs = getFormula(pf);

      ASSERT(lhs->getType() == POS_LITERAL);
      ASSERT(!rhs->checkOccurence(lhs));

      string* sym = ((FormulaLiteral*) lhs)->getSymbol();
      ASSERT(conceptDefs.find(sym) == conceptDefs.end());
      conceptDefs[sym] = rhs;
      lastConceptName = sym; 

      s = getSym(pf); 
      ASSERT(s == SYM_PAR_RIGHT);
      s = getSym(pf);
      continue;
    };
  }; // while
};

void ALCInOut::display(Formula* f, FILE *pf)
{
  char type = f->getType();
  switch (type) {
    case TOP: fputs("*TOP*", pf); break;
    case BOT: fputs("*BOTTOM*", pf); break;
    case POS_LITERAL: fputs(((FormulaLiteral*) f)->getSymbol()->c_str(), pf); break;
    case NEG_LITERAL: fprintf(pf, "(NOT %s)", ((FormulaLiteral*) f)->getSymbol()->c_str()); break;

    case FORALL: 
    case EXISTS: 
      fprintf(pf, type == FORALL ? "(ALL " : "(SOME ");
      display(((FormulaModalQuant*) f)->getModalIndex(), pf);
      fputs(" ", pf);
      display(((FormulaModalQuant*) f)->getQuantFormula(), pf);
      fputs(")", pf);
      break;

    case AND_SET:
    case OR_SET:
    {
      set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();
      set<Formula*>::iterator it = fs->begin();
      int count = fs->size();
      while (count >= 2) {
        fputs(type == AND_SET ? "(AND " : "(OR ", pf);
        display(*it, pf);
        fputs(" ", pf);
        it++;
        count--;
      };
      display(*it, pf);
      for (int i = fs->size(); i >= 2; i--)
        fputs(")", pf);
      break;
    };
  }; // switch
};

Formula* ALCInOut::getConceptDef(string conceptName)
{
  string* sym = SymbolCatalogue::find(conceptName);
  map<string*,Formula*>::iterator it = conceptDefs.find(sym);
  if (it != conceptDefs.end())
    return it->second;

  return FormulaCatalogue::find(new FormulaLiteral(POS_LITERAL, sym));
};
