/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
using namespace std;

#include <map>
#include <set>
#include <stack>
#include <math.h>

#include "options.h"
#include "statistics.h"
#include "ALC.h"

bool ALCNode::chooseRule(labels_queue &labels) // updates rule, principal, labels, otherInfo
{
  if (Options::literalElimination || Options::semanticBranching1 || Options::semanticBranching2) {

    if (Options::literalElimination)
      STATISTICS_START(eliminatingLiterals);
    if (Options::semanticBranching1)
      STATISTICS_START(semanticBranching1);
    if (Options::semanticBranching2)
      STATISTICS_START(semanticBranching2);

    map<Formula*,int> pos, neg, box, dia;
    content->countOccurrencesOfAtoms(pos, neg, box, dia);

    if (Options::literalElimination)
      STATISTICS_STOP(eliminatingLiterals);
    if (Options::semanticBranching1)
      STATISTICS_STOP(semanticBranching1);
    if (Options::semanticBranching2)
      STATISTICS_STOP(semanticBranching2);

    if (Options::literalElimination) {
      STATISTICS_START(eliminatingLiterals);
      int count = 0;

      otherInfo = new NodeInfo_ParamsOfLE();
      set<Formula*> *pureLiterals    = &( ((NodeInfo_ParamsOfLE *) otherInfo)->pureLiterals );
      set<Formula*> *clsUnitFormulas = &( ((NodeInfo_ParamsOfLE *) otherInfo)->clsUnitFormulas );
      set<Formula*> *modUnitFormulas = &( ((NodeInfo_ParamsOfLE *) otherInfo)->modUnitFormulas );

      for (map<Formula*,int>::iterator it = pos.begin(); it != pos.end(); it++)
        if (neg.find(it->first) == neg.end()) {
          pureLiterals->insert(it->first);
          count += it->second;
        };

      for (map<Formula*,int>::iterator it = neg.begin(); it != neg.end(); it++)
        if (pos.find(it->first) == pos.end()) {
          pureLiterals->insert(it->first->getNegation());
          count += it->second;
        };

      if (content->getType() == AND_SET) {
        set<Formula*> *fs = ((FormulaSet *) content)->getSetOfFormulas();
        for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++) {
          if ((*it)->getType() == POS_LITERAL || (*it)->getType() == NEG_LITERAL) {
            if (pureLiterals->find(*it) == pureLiterals->end()) {
              clsUnitFormulas->insert(*it);
              if ((*it)->getType() == POS_LITERAL)
                count += pos[*it] + neg[*it];
              else
                count += pos[(*it)->getNegation()] + neg[(*it)->getNegation()];
            };
          }
          else if ((*it)->getType() == FORALL) {
            int k = box[*it] + dia[(*it)->getNegation()];
            if (k > 1) {
              modUnitFormulas->insert(*it);
              count += k-1;
            };
          }
          else if ((*it)->getType() == EXISTS) {
            int k = dia[*it] + box[(*it)->getNegation()];
            if (k > 1) {
              modUnitFormulas->insert(*it);
              count += k-1;
            };
          };
        }; // for
      }; // if

      if (count >= Options::literalEliminationThreshold) {
        rule = RULE_LIT_ELIM;
        principal = NULL;
        labels.push(FormulaCatalogue::top);
        STATISTICS_STOP(eliminatingLiterals);
        #ifdef TESTING
        cout << "Eliminating literals: " << count << " occurences ";
        #endif 
        return true;
      }
      else {
        delete otherInfo;
        otherInfo = NULL;
        STATISTICS_STOP(eliminatingLiterals);
      };
    }; // if (Options::literalElimination)

    if (Options::semanticBranching1 || Options::semanticBranching2) {

      Formula* atom1 = NULL;
      Formula* atom2 = NULL;
      int numOfOccurences1 = 0;
      int numOfOccurences2 = 0;

      if (Options::semanticBranching1) {
        STATISTICS_START(semanticBranching1);

        for (map<Formula*,int>::iterator it = neg.begin(); it != neg.end(); it++)
          pos[it->first] += it->second;


        for (map<Formula*,int>::iterator it = pos.begin(); it != pos.end(); it++)
          if (it->second > numOfOccurences1) {
            atom1 = it->first;
            numOfOccurences1 = it->second;
          };

        STATISTICS_STOP(semanticBranching1);
      }; // if

      if (Options::semanticBranching2) {
        STATISTICS_START(semanticBranching2);

        for (map<Formula*,int>::iterator it = dia.begin(); it != dia.end(); it++)
          box[it->first->getNegation()] += it->second;

        double size_or_weight = 0;

        for (map<Formula*,int>::iterator it = box.begin(); it != box.end(); it++) {
          if (it->second > numOfOccurences2) {
            atom2 = it->first;
            numOfOccurences2 = it->second;
            size_or_weight = (Options::usingSize4OrBranching ? 
		atom2->getSize() : atom2->getWeight());
          }
          else if (it->second == numOfOccurences2) {
            double x = (Options::usingSize4OrBranching ? 
	 		it->first->getSize() : it->first->getWeight());
            if (x > size_or_weight) {
              atom2 = it->first;
              numOfOccurences2 = it->second;
              size_or_weight = x;
            };
          }; //else
        }; // for

        STATISTICS_STOP(semanticBranching2);
      }; // if

      if (numOfOccurences1 >= Options::semanticBranching1Threshold)
        if (numOfOccurences2 < Options::semanticBranching2Threshold ||
            numOfOccurences1 >= numOfOccurences2 - 1) {
          rule = RULE_SEM_BRANCHING1;
          principal = atom1;
          labels.push(FormulaCatalogue::top);
          labels.push(FormulaCatalogue::bot);
          return true;
        }; // if

      if (numOfOccurences2 >= Options::semanticBranching2Threshold) {
        rule = RULE_SEM_BRANCHING2;
        principal = atom2;
        labels.push(FormulaCatalogue::top);
        labels.push(FormulaCatalogue::bot);
        return true;
      };
    }; // if (Options::semanticBranching1 || Options::semanticBranching2)
  }; // if

  if (Options::simplifyNodes) {
    if (mainParent) 
      if (mainParent->getRule() != RULE_SIMP) {
        Formula* tauContent = Graph::graph->tau->getContent();
        double p = ((double) content->getSize()) / tauContent->getSize();
        if (p <= Options::relativeNodeSize4Simp) {
          STATISTICS_START(simplifyingFormulas);
          Formula* f = content->sat_simplify(Graph::graph->unsatFormulas);
          STATISTICS_STOP(simplifyingFormulas);
          if (f->getSize() < content->getSize()) {
            rule = RULE_SIMP;
            principal = NULL;
            labels.push(FormulaCatalogue::top);
            otherInfo = new NodeInfo_4SR(f);
            return true;
          };
        };
      };
  };

  if (content->getType() == OR_SET) {
    rule = RULE_OR;
    principal = content;
    set<Formula*> *fs = ((FormulaSet *) content)->getSetOfFormulas();
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
      labels.push(*it);
    return true;
  }

  if (content->getType() == EXISTS) {
    rule = RULE_TRANS;
    principal = NULL;
    labels.push(content);
    return true;
  };

  if (content->getType() == AND_SET) {
    set<Formula*> *fs = ((FormulaSet *) content)->getSetOfFormulas();
    labels_queue labels2;

/* old:
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
      if ((*it)->getType() == OR_SET) {
        rule = RULE_OR;
        principal = *it;
        set<Formula*> *fs2 = ((FormulaSet *) *it)->getSetOfFormulas();
        for (set<Formula*>::iterator it2 = fs2->begin(); it2 != fs2->end(); it2++)
          labels.push(*it2);
        return true;
      }
      else if ((*it)->getType() == EXISTS)
        labels2.push(*it);
*/

// new:
    set<Formula*>::iterator it0 = fs->end();
    int maxPrincipalSize = 0;
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
      if ((*it)->getType() == OR_SET) {
        if (Options::usingSize4OrBranching) {
          if ((*it)->getSize() > maxPrincipalSize) {
            maxPrincipalSize = (*it)->getSize();
            it0 = it;
          }
        }
        else
          if ((*it)->getWeight() > maxPrincipalSize) {
            maxPrincipalSize = (int) trunc((*it)->getWeight());
            it0 = it;
          }
      }
      else if ((*it)->getType() == EXISTS)
        labels2.push(*it);

    if (it0 != fs->end()){
      rule = RULE_OR;
      principal = *it0;
      set<Formula*> *fs2 = ((FormulaSet *) *it0)->getSetOfFormulas();
      for (set<Formula*>::iterator it2 = fs2->begin(); it2 != fs2->end(); it2++)
        labels.push(*it2);
      return true;
    }
//

    if (!labels2.empty()) {
      rule = RULE_TRANS;
      principal = NULL;
      labels = labels2;
      return true;
    };
  };

  return false;
};

Formula* ALCNode::applyRule(Formula* label, Formula* generalAssumptions)
{
  if (rule == RULE_LIT_ELIM) {
    STATISTICS_START(eliminatingLiterals);

    NodeInfo_ParamsOfLE *paramsOfLE = (NodeInfo_ParamsOfLE*) otherInfo;
    set<Formula*> literals;
    literals.insert(paramsOfLE->pureLiterals.begin(), paramsOfLE->pureLiterals.end());
    literals.insert(paramsOfLE->clsUnitFormulas.begin(), paramsOfLE->clsUnitFormulas.end());
    literals.insert(paramsOfLE->modUnitFormulas.begin(), paramsOfLE->modUnitFormulas.end());

    #ifdef TESTING
    cout << " (" 
	<< paramsOfLE->pureLiterals.size() << ", " 
	<< paramsOfLE->clsUnitFormulas.size() << ", " 
	<< paramsOfLE->modUnitFormulas.size() << ")\n";
    #endif 

    if (content->getType() != AND_SET) {
      Formula* f = content->replaceAt1Level(literals);
      otherInfo = NULL;
      if (!paramsOfLE->modUnitFormulas.size()) {
        delete paramsOfLE;
        STATISTICS_STOP(eliminatingLiterals);
        return f;
      }
      else {
        delete paramsOfLE;
        FormulaSet *f2 = new FormulaSet(AND_SET);
        set<Formula*> *fs2 = f2->getSetOfFormulas();
        fs2->insert(paramsOfLE->modUnitFormulas.begin(), paramsOfLE->modUnitFormulas.end());
        f2->add(f, Options::checkSubsumption4FirstLayer);

        Formula* f3;
        f2->findSimpleForm(f3);
        STATISTICS_STOP(eliminatingLiterals);
        if (f3) { delete f2; return f3; }
        else return FormulaCatalogue::find(f2);
      }; // else
    }
    else {
      otherInfo = new NodeInfo_CauseMap4LE();
      map<Formula*,Formula*> *cmap = &( ((NodeInfo_CauseMap4LE*) otherInfo)->causeMap );
      set<Formula*> *unitFormulas = NULL;
      if (paramsOfLE->clsUnitFormulas.size() || paramsOfLE->modUnitFormulas.size()) {
        ((NodeInfo_CauseMap4LE*) otherInfo)->unitFormulas = new set<Formula*>;
        unitFormulas = ((NodeInfo_CauseMap4LE*) otherInfo)->unitFormulas;
        unitFormulas->insert(paramsOfLE->clsUnitFormulas.begin(),
			     paramsOfLE->clsUnitFormulas.end());
        unitFormulas->insert(paramsOfLE->modUnitFormulas.begin(),
			     paramsOfLE->modUnitFormulas.end());
      };

      FormulaSet *f2 = new FormulaSet(AND_SET);
      set<Formula*> *fs2 = f2->getSetOfFormulas();
      fs2->insert(paramsOfLE->modUnitFormulas.begin(), paramsOfLE->modUnitFormulas.end());
      delete paramsOfLE;

      set<Formula*> *content_fs = ((FormulaSet*) content)->getSetOfFormulas();
      for (set<Formula*>::iterator it = content_fs->begin(); it != content_fs->end(); it++) {
        Formula* g = (*it)->replaceAt1Level(literals);
        if (g == FormulaCatalogue::bot) {
          (*cmap)[g] = *it;
          delete f2;
          STATISTICS_STOP(eliminatingLiterals);
          return FormulaCatalogue::bot;
        }
        else if (g == FormulaCatalogue::top)
          continue;
        else {
          f2->add(g, Options::checkSubsumption4FirstLayer);

          if (g->getType() != AND_SET) {
            if (fs2->find(g) != fs2->end())
              (*cmap)[g] = *it;
          }
          else {
            set<Formula*> *gs = ((FormulaSet*) g)->getSetOfFormulas();
            for (set<Formula*>::iterator it2 = gs->begin(); it2 != gs->end(); it2++)
              if (fs2->find(*it2) != fs2->end())
                (*cmap)[*it2] = *it;
          };
        }; // else
      }; // for

      if (fs2->empty()) {
        delete f2;
        STATISTICS_STOP(eliminatingLiterals);
        return FormulaCatalogue::top;
      };

      Formula* f3;
      f2->findSimpleForm(f3);
      STATISTICS_STOP(eliminatingLiterals);
      if (f3) { delete f2; return f3; }
      else return FormulaCatalogue::find(f2);
    }; // else
  }; // if (rule == RULE_LIT_ELIM)


  if (rule == RULE_SEM_BRANCHING1) {
    STATISTICS_START(semanticBranching1);
    assert(label == FormulaCatalogue::top || label == FormulaCatalogue::bot);
    bool val = (label == FormulaCatalogue::top);

    if (content->getType() != AND_SET) {
      Formula* f = content->replaceAt1Level(principal, val);
      STATISTICS_STOP(semanticBranching1);
      return f;
    }
    else {
      if (!otherInfo)
        otherInfo = new NodeInfo_CauseMaps4SB();
      NodeInfo_CauseMaps4SB* cmap2 = (NodeInfo_CauseMaps4SB*) otherInfo;
      map<Formula*,Formula*> *cmap = (val ? &(cmap2->byPosLit) : &(cmap2->byNegLit));

      FormulaSet *f2 = new FormulaSet(AND_SET);
      set<Formula*> *fs2 = f2->getSetOfFormulas();

      set<Formula*> *content_fs = ((FormulaSet*) content)->getSetOfFormulas();
      for (set<Formula*>::iterator it = content_fs->begin(); it != content_fs->end(); it++) {
        Formula* g = (*it)->replaceAt1Level(principal, val);
        if (g == FormulaCatalogue::bot) {
          (*cmap)[g] = *it;
          delete f2;
          STATISTICS_STOP(semanticBranching1);
          return FormulaCatalogue::bot;
        }
        else if (g == FormulaCatalogue::top)
          continue;
        else {
          f2->add(g, Options::checkSubsumption4FirstLayer);

          if (g->getType() != AND_SET) {
            if (fs2->find(g) != fs2->end())
              (*cmap)[g] = *it;
          }
          else {
            set<Formula*> *gs = ((FormulaSet*) g)->getSetOfFormulas();
            for (set<Formula*>::iterator it2 = gs->begin(); it2 != gs->end(); it2++)
              if (fs2->find(*it2) != fs2->end())
                (*cmap)[*it2] = *it;
          };
        }; // else
      }; // for

      if (fs2->empty()) {
        delete f2;
        STATISTICS_STOP(semanticBranching1);
        return FormulaCatalogue::top;
      };

      Formula* f3;
      f2->findSimpleForm(f3);
      STATISTICS_STOP(semanticBranching1);
      if (f3) { delete f2; return f3; }
      else return FormulaCatalogue::find(f2);
    }; // else
  }; // if (rule == RULE_SEM_BRANCHING1)


  if (rule == RULE_SEM_BRANCHING2) {
    STATISTICS_START(semanticBranching2);
    assert(label == FormulaCatalogue::top || label == FormulaCatalogue::bot);
    bool val = (label == FormulaCatalogue::top);

    if (content->getType() != AND_SET) {
      Formula* f = content->replaceAt1Level(principal, val);
      FormulaSet g(AND_SET, f, val ? principal : principal->getNegation());
      STATISTICS_STOP(semanticBranching2);
      return g.normalize();
    }
    else {
      if (!otherInfo)
        otherInfo = new NodeInfo_CauseMaps4SB();
      NodeInfo_CauseMaps4SB* cmap2 = (NodeInfo_CauseMaps4SB*) otherInfo;
      map<Formula*,Formula*> *cmap = (val ? &(cmap2->byPosLit) : &(cmap2->byNegLit));

      FormulaSet *f2 = new FormulaSet(AND_SET);
      set<Formula*> *fs2 = f2->getSetOfFormulas();
      assert(principal->getType() == FORALL || principal->getType() == EXISTS);
      fs2->insert(val ? principal : principal->getNegation());

      set<Formula*> *content_fs = ((FormulaSet*) content)->getSetOfFormulas();
      for (set<Formula*>::iterator it = content_fs->begin(); it != content_fs->end(); it++) {
        Formula* g = (*it)->replaceAt1Level(principal, val);
        if (g == FormulaCatalogue::bot) {
          (*cmap)[g] = *it;
          delete f2;
          STATISTICS_STOP(semanticBranching2);
          return FormulaCatalogue::bot;
        }
        else if (g == FormulaCatalogue::top)
          continue;
        else {
          f2->add(g, Options::checkSubsumption4FirstLayer);

          if (g->getType() != AND_SET) {
            if (fs2->find(g) != fs2->end())
              (*cmap)[g] = *it;
          }
          else {
            set<Formula*> *gs = ((FormulaSet*) g)->getSetOfFormulas();
            for (set<Formula*>::iterator it2 = gs->begin(); it2 != gs->end(); it2++)
              if (fs2->find(*it2) != fs2->end())
                (*cmap)[*it2] = *it;
          };
        }; // else
      }; // for

      if (fs2->empty()) {
        delete f2;
        STATISTICS_STOP(semanticBranching2);
        return FormulaCatalogue::top;
      };

      Formula* f3;
      f2->findSimpleForm(f3);
      STATISTICS_STOP(semanticBranching2);
      if (f3) { delete f2; return f3; }
      else return FormulaCatalogue::find(f2);
    }; // else
  }; // if (rule == RULE_SEM_BRANCHING2)


  if (rule == RULE_SIMP) {
    Formula* f = ((NodeInfo_4SR*) otherInfo)->simplifiedFormula;
    delete otherInfo;
    otherInfo = NULL;
    return f;
  };

  if (rule == RULE_OR) {
    if (content == principal)
      return label;
    else { // content->getType() == AND_SET
      FormulaSet *f2 = new FormulaSet((FormulaSet&) *content);
      f2->erase(principal);
      f2->add(label, Options::checkSubsumption4FirstLayer);

      Formula* f3;
      f2->findSimpleForm(f3);
      if (f3) { delete f2; return f3; }
      else return FormulaCatalogue::find(f2);
    };
  }
  else {
    assert(!principal && rule == RULE_TRANS && label->getType() == EXISTS);

    ModalIndex* mIdx = ((FormulaModalQuant*) label)->getModalIndex();

    FormulaSet *f2 = new FormulaSet(AND_SET);
    f2->add( ((FormulaModalQuant*) label)->getQuantFormula() );

    if (content->getType() == AND_SET) {
      set<Formula*> *fs = ((FormulaSet *) content)->getSetOfFormulas();
      for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++) {
        if ((*it)->getType() == FORALL) {
          if (((FormulaModalQuant*) *it)->getModalIndex() == mIdx) {
            f2->add( ((FormulaModalQuant*) *it)->getQuantFormula(),
                     Options::checkSubsumption4FirstLayer );
          };
        };
      };
    };

    Formula* f3;
    f2->findSimpleForm(f3);
    if (f3) { delete f2; return f3; }
    else return FormulaCatalogue::find(f2);
  };
};

Formula* ALCNode::computeUnsatCore(Formula* generalAssumptions, Node* child)
{
  assert(Options::useUnsatCores);

  if (content->getType() != AND_SET)
    return content;

  if (rule == RULE_SIMP)
    return content;

  set<Formula*> *content_fs = ((FormulaSet*) content)->getSetOfFormulas();


  if (rule == RULE_LIT_ELIM) {
    STATISTICS_START(eliminatingLiterals);
    assert(kind == AND_NODE && child);
    assert(otherInfo);

    map<Formula*,Formula*> *cm   = &( ((NodeInfo_CauseMap4LE*) otherInfo)->causeMap ); 
    set<Formula*> *unitFormulas = ((NodeInfo_CauseMap4LE*) otherInfo)->unitFormulas;

    FormulaSet *f = new FormulaSet(AND_SET);
    if (unitFormulas)
      f->getSetOfFormulas()->insert(unitFormulas->begin(), unitFormulas->end());

    Formula* unsatChildContent = child->getContent();
    if (unsatChildContent->getType() != AND_SET) {
      bool bCheck = true;
      if (unitFormulas)
        bCheck = (unitFormulas->find(unsatChildContent) == unitFormulas->end());
      if (bCheck) {
        map<Formula*,Formula*>::iterator it = cm->find(unsatChildContent);
        assert(it != cm->end());
        assert(content_fs->find(it->second) != content_fs->end());
        f->insert(it->second);
      };
    }
    else {
      set<Formula*> *cfs = ((FormulaSet*) unsatChildContent)->getSetOfFormulas();
      for (set<Formula*>::iterator it = cfs->begin(); it != cfs->end(); it++) {
        bool bCheck = true;
        if (unitFormulas)
          bCheck = (unitFormulas->find(*it) == unitFormulas->end());
        if (bCheck) {
          map<Formula*,Formula*>::iterator it2 = cm->find(*it);
          assert(it2 != cm->end());
          assert(content_fs->find(it2->second) != content_fs->end());
          f->insert(it2->second);
        }; // if
      }; // for
    }; // else

    delete otherInfo;
    otherInfo = NULL;

    Formula* _f;
    f->findSimpleForm(_f);
    STATISTICS_STOP(eliminatingLiterals);
    if (_f) { delete f; return _f; }
    else return FormulaCatalogue::find(f);
  }; // if (_rule == RULE_LIT_ELIM)


  if (rule == RULE_SEM_BRANCHING1) {
    STATISTICS_START(semanticBranching1);
    assert(kind == OR_NODE && !child);
    assert(otherInfo);

    NodeInfo_CauseMaps4SB* cm2 = (NodeInfo_CauseMaps4SB*) otherInfo; 

    FormulaSet *f = new FormulaSet(AND_SET);
    for (map<Node*,Formula*>::iterator it0 = edges.begin(); it0 != edges.end(); it0++) {
      child = it0->first;
      map<Formula*,Formula*> *cm = (it0->second == FormulaCatalogue::top ? 
                                    &(cm2->byPosLit) : &(cm2->byNegLit));

      Formula* unsatChildContent = child->getContent();
      if (unsatChildContent->getType() != AND_SET) {
        map<Formula*,Formula*>::iterator it = cm->find(unsatChildContent);
        assert(it != cm->end());
        assert(content_fs->find(it->second) != content_fs->end());
        f->insert(it->second);
      }
      else {
        set<Formula*> *cfs = ((FormulaSet*) unsatChildContent)->getSetOfFormulas();
        for (set<Formula*>::iterator it = cfs->begin(); it != cfs->end(); it++) {
          map<Formula*,Formula*>::iterator it2 = cm->find(*it);
          assert(it2 != cm->end());
          assert(content_fs->find(it2->second) != content_fs->end());
          f->insert(it2->second);
        };
      };
    }; // for

    delete otherInfo;
    otherInfo = NULL;

    Formula* _f;
    f->findSimpleForm(_f);
    STATISTICS_STOP(semanticBranching1);
    if (_f) { delete f; return _f; }
    else return FormulaCatalogue::find(f);
  }; // if (rule == RULE_SEM_BRANCHING1)


  if (rule == RULE_SEM_BRANCHING2) {
    STATISTICS_START(semanticBranching2);
    assert(kind == OR_NODE && !child);
    assert(otherInfo);

    NodeInfo_CauseMaps4SB* cm2 = (NodeInfo_CauseMaps4SB*) otherInfo; 

    FormulaSet *f = new FormulaSet(AND_SET);
    for (map<Node*,Formula*>::iterator it0 = edges.begin(); it0 != edges.end(); it0++) {
      child = it0->first;
      bool val = (it0->second == FormulaCatalogue::top);
      map<Formula*,Formula*> *cm = (it0->second == FormulaCatalogue::top ? 
                                    &(cm2->byPosLit) : &(cm2->byNegLit));

      Formula* unsatChildContent = child->getContent();
      if (unsatChildContent->getType() != AND_SET) {
        if (val ? unsatChildContent != principal : 
	          unsatChildContent != principal->getNegation()) {
          map<Formula*,Formula*>::iterator it = cm->find(unsatChildContent);
          assert(it != cm->end());
          assert(content_fs->find(it->second) != content_fs->end());
          f->insert(it->second);
        };
      }
      else {
        set<Formula*> *cfs = ((FormulaSet*) unsatChildContent)->getSetOfFormulas();
        for (set<Formula*>::iterator it = cfs->begin(); it != cfs->end(); it++) {
          if (val ? *it != principal : *it != principal->getNegation()) {
            map<Formula*,Formula*>::iterator it2 = cm->find(*it);
            assert(it2 != cm->end());
            assert(content_fs->find(it2->second) != content_fs->end());
            f->insert(it2->second);
          };
        };
      };
    }; // for

    delete otherInfo;
    otherInfo = NULL;

    Formula* _f;
    f->findSimpleForm(_f);
    STATISTICS_STOP(semanticBranching2);
    if (_f) { delete f; return _f; }
    else return FormulaCatalogue::find(f);
  }; // if (rule == RULE_SEM_BRANCHING2)


  if (rule == RULE_TRANS) {
    assert(kind == AND_NODE && child);

    FormulaSet *f = new FormulaSet((FormulaSet&) *content);
    set<Formula*> *fs = f->getSetOfFormulas();

    Formula* label = edges[child];
    assert(label);
    assert(label->getType() == EXISTS);

    set<Formula*>::iterator _it;
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); )
      if (*it != label) {
        if ((*it)->getType() != FORALL) {
          _it = it++;
          f->erase(_it);
        }
        else if (((FormulaModalQuant*) *it)->getModalIndex() != 
                 ((FormulaModalQuant*) label)->getModalIndex()) {
          _it = it++;
          f->erase(_it);
        }
        else it++;
      }
      else it++;

    Formula* unsatChildContent = child->getContent();
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); ) {
      if ((*it)->getType() != FORALL) it++;
      else {
        Formula* g = ((FormulaModalQuant*) *it)->getQuantFormula();
        if (g == unsatChildContent)
          { it++; continue; };
          assert(g->getType() != AND_SET);
          if (unsatChildContent->getType() == AND_SET) {
            set<Formula*> *cfs = ((FormulaSet*) unsatChildContent)->getSetOfFormulas();
            if (cfs->find(g) != cfs->end())
              { it++; continue; };
          };
          _it = it++;
          f->erase(_it);
      }; // else
    };  // for

    Formula* _f;
    f->findSimpleForm(_f);
    if (_f) { delete f; return _f; }
    else return FormulaCatalogue::find(f);
  }; // if (rule == RULE_TRANS)


  if (rule == RULE_OR) {
    assert(kind == OR_NODE && !child);

    FormulaSet *f = new FormulaSet(AND_SET);

    for (map<Node*,Formula*>::iterator it = edges.begin(); it != edges.end(); it++) {
      child = it->first;
      Formula* label = it->second;
      if (label == child->getContent())
        continue;
      if (child->getContent()->getType() != AND_SET) {
        f->add(child->getContent());
        continue;
      };
      set<Formula*> gs(*((FormulaSet*) child->getContent())->getSetOfFormulas());
      gs.erase(label);
      for (set<Formula*>::iterator it2 = gs.begin(); it2 != gs.end(); it2++)
        f->add(*it2);
    };

    if (!f->getSetOfFormulas()->size()) {
      delete f;
      return principal;
    };

    set<Formula*> *fs = f->getSetOfFormulas();
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); )
      if (content_fs->find(*it) == content_fs->end()) {
        set<Formula*>::iterator _it = it++;
        f->erase(_it);
      }
      else it++;
    f->add(principal);

    Formula* _f;
    f->findSimpleForm(_f);
    if (_f) { delete f; return _f; }
    else return FormulaCatalogue::find(f);
  }; // if (rule == RULE_OR)
};
