/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _CAT_H
#define _CAT_H

#include <string>
#include <map>

#include "first.h"
#include "formula0.h"
#include "hashtable.h"

//-----------------------------------------------------------------------------

class SymbolCatalogue {
  private:
    static map<string,string*> symbols;
  public:
    static string* find(string &s); 
};

//-----------------------------------------------------------------------------
// hashtable-support was implemented by Lukasz Czajka

#ifdef HASHTABLE_SUPPORT

class FormulaCatalogue {
  public:
    typedef Hashtable<Formula,FormulaHasher,FormulaEqCmp> hashtab_t;

  private:
    static hashtab_t formulas;

  public:
    static Formula* top;
    static Formula* bot;

    static void init();
    static Formula* find0(Formula *f) { return formulas.find_elem(f); } 
    static Formula* find(Formula *f); 
    static Formula* insert(Formula *f) { formulas.insert(f); return f; }
    static void reduce();
    static void clear();
    static void checkAndReduce();

    // other methods:
    static int getSize() { return formulas.size(); };
    static hashtab_t* getFormulas() { return &formulas; };

    static void reserve(unsigned size) { formulas.reserve(size); };
    static void preallocate(int num);

    // for testing:
    static int getSize2();
    static int getNumOfCollisions();
    void display(FILE *f);
};

#else 

//-----------------------------------------------------------------------------
// without hashtable support

class FormulaCatalogue {
  private:
    static multimap<uint32_t,Formula*> formulas;

  public:
    static Formula* top;
    static Formula* bot;

    // main methods:
    static void init();
    static Formula* find0(Formula *f); // find without adding
    static Formula* find(Formula *f); // find and add if unfound; *f is used by the function
    static Formula* insert(Formula *f);
    static void reduce();
    static void clear();
    static void checkAndReduce();

    // other methods:
    static int getSize() { return formulas.size(); };
    static multimap<uint32_t,Formula*>* getFormulas() { return &formulas; };
    static void preallocate(int num) {};

    // for testing:
    static int getSize2(); // couting also length of AND_SET and OR_SET formulas
    static int getNumOfCollisions();
    void display(FILE *f); 
};

#endif // hashtable support

#endif
