/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include <stdlib.h>
#include <sys/types.h>
//#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <unistd.h>

#include "tools.h"
#include "concurency.h"
#include "statistics.h"

static int shmid;

void sigterm(int sig)
{
  STATISTICS_STOP(totalTime);
  STATISTICS_SET(exited, true);

  shmdt((void*) _stat);
  struct shmid_ds buf;
  shmctl(shmid, IPC_RMID, &buf);
  exit(1);
};

void allocChildInfo()
{
  key_t key;
  struct shmid_ds buf;

  srand((int) clock());
  shmid = -1;
  for (int i = 0; i < 1000; i++) {
    key = (key_t) rand();
    if ((shmid = shmget(key, sizeof(ChildInfo), IPC_CREAT | 0600)) != -1)
      break;
  };
  srand(1);

  if (shmid == -1) {
    MSG("Attempts of calling \"shmget\" failed.");
    exit(-1);
  };

  _stat = (ChildInfo*) shmat(shmid, 0, 0);
  if (_stat == (ChildInfo*)(-1)) {
    MSG("shmat failure.");
    shmctl(shmid, IPC_RMID, &buf);
    exit(-1);
  };

  if (signal(SIGTERM, &sigterm) == SIG_ERR) {
    MSG("signal failure.");
    shmctl(shmid, IPC_RMID, &buf);
    exit(-1);
  };
};

void deallocChildInfo()
{
  if (shmdt((void*) _stat) == -1)
    DEBUG_MSG("shmdt failure.");

  struct shmid_ds buf;
  if (shmctl(shmid, IPC_RMID, &buf) == -1)
    DEBUG_MSG("shmctl failure.");
};

void sig_alarm_for_child(int sig)
{
  STATISTICS_STOP(totalTime);
  STATISTICS_SET(exited, true);
  DEBUG_MSG("Runs out of time.");
  exit(1);
};

void scheduleExitTimeForChild(int seconds)
{
  if (signal(SIGALRM, &sig_alarm_for_child) == SIG_ERR) {
    MSG("signal failure.");
    exit(-1);
  };

  alarm(seconds);
};
