/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include "options.h"
#include "formula.h"

//------------------------------------------------------------------------------

#define usual_check_for_subsumption() { \
  if (f == this) \
    return true; \
  \
  if (flag) \
    if (unsats.find(getNegation()) != unsats.end()) \
      return true; \
  \
  if (f->getType() == AND_SET) \
    if (isSubsumedBy(*((FormulaSet*) f)->getSetOfFormulas(), unsats, false)) \
      return true; \
  \
  if (f->getType() == OR_SET) { \
    bool ok = true; \
    set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas(); \
    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++) \
      if (!isSubsumedBy(*it, unsats, false)) { ok = false; break; }; \
    if (ok) \
      return true; \
  }; \
}

bool FormulaLiteral::isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag)
{
  usual_check_for_subsumption();
  return false;
};

bool FormulaModalQuant::isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag)
{
  usual_check_for_subsumption();

  if (f->getType() == type)
    if (((FormulaModalQuant*) f)->getModalIndex() == modalIndex)
      if (qFormula->isSubsumedBy(((FormulaModalQuant*) f)->getQuantFormula(), unsats, true))
        return true;

  return false;
};

bool FormulaSet::isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag)
{
  if (f == this)
    return true;

  if (flag)
    if (unsats.find(getNegation()) != unsats.end())
      return true;

  if (type == AND_SET) {
    for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
      if (!(*it)->isSubsumedBy(f, unsats, true))
        return false;

    return true;
  }
  else { // type == OR_SET
    for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
      if ((*it)->isSubsumedBy(f, unsats, true))
        return true;

    if (isSuperSetOf(OR_SET, f))
      return true;

    for (set<Formula*>::iterator it = unsats.begin(); it != unsats.end(); it++)
      if (getNegation()->isSuperSetOf(AND_SET, *it))
        return true;

    return false;
  };
};

//------------------------------------------------------------------------------

#define usual_check_for_subsumption2() { \
  if (local.find(this) != local.end()) \
    return true; \
  \
  if (flag) \
    if (unsats.find(getNegation()) != unsats.end()) \
      return true; \
  \
  for (set<Formula*>::iterator it = local.begin(); it != local.end(); it++) \
    if (isSubsumedBy(*it, unsats, false)) \
      return true; \
}

bool FormulaLiteral::isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag)
{
  usual_check_for_subsumption2();
  return false;
};

bool FormulaModalQuant::isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag)
{
  usual_check_for_subsumption2();

  set<Formula*> local2;
  for (set<Formula*>::iterator it = local.begin(); it != local.end(); it++)
    if ((*it)->getType() == FORALL)
      if (((FormulaModalQuant*) *it)->getModalIndex() == modalIndex)
        local2.insert(((FormulaModalQuant*) *it)->getQuantFormula());

  if (local2.size() > 1) {
    if (type == FORALL) {
      if (qFormula->isSubsumedBy(local2, unsats, true))
        return true;
    }
    else {  // type == EXISTS
      for (set<Formula*>::iterator it = local.begin(); it != local.end(); it++)
        if ((*it)->getType() == EXISTS)
          if (((FormulaModalQuant*) *it)->getModalIndex() == modalIndex) {
            set<Formula*> local3(local2);
            local3.insert(((FormulaModalQuant*) *it)->getQuantFormula());
            if (local3.size() > 1)
              if (qFormula->isSubsumedBy(local3, unsats, true))
                return true;
          };
    }; // else
  }; // if

  return false;
};

bool FormulaSet::isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag)
{
  if (local.find(this) != local.end())
    return true;

  if (flag)
    if (unsats.find(getNegation()) != unsats.end())
      return true;

  if (type == AND_SET) {
    for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
      if (!(*it)->isSubsumedBy(local, unsats, true))
        return false;

    return true;
  }
  else { // type == OR_SET
    for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
      if ((*it)->isSubsumedBy(local, unsats, true))
        return true;

    for (set<Formula*>::iterator it = local.begin(); it != local.end(); it++)
      if (isSuperSetOf(OR_SET, *it))
        return true;

    for (set<Formula*>::iterator it = unsats.begin(); it != unsats.end(); it++)
      if (getNegation()->isSuperSetOf(AND_SET, *it))
        return true;

    return false;
  };
};

//------------------------------------------------------------------------------

Formula* FormulaLiteral::simplify(set<Formula*> &local, set<Formula*> &unsats)
{
  if (isSubsumedBy(local, unsats))
    return FormulaCatalogue::top;

  if (getNegation()->isSubsumedBy(local, unsats))
    return FormulaCatalogue::bot;

  return this;
};

Formula* FormulaModalQuant::simplify(set<Formula*> &local, set<Formula*> &unsats)
{
  set<Formula*> local2;
  Formula* f = qFormula->simplify(local2, unsats);

  if (type == FORALL && f == FormulaCatalogue::top)
    return FormulaCatalogue::top;

  if (type == EXISTS && f == FormulaCatalogue::bot)
    return FormulaCatalogue::bot;

  if (f == qFormula) { 
    if (isSubsumedBy(local, unsats))
      return FormulaCatalogue::top;
    if (getNegation()->isSubsumedBy(local, unsats))
      return FormulaCatalogue::bot;
    return this;
  }
  else {
    FormulaModalQuant g(type, modalIndex, f);
    Formula* h = g.normalize();

    if (h->isSubsumedBy(local, unsats))
      return FormulaCatalogue::top;
    if (h->getNegation()->isSubsumedBy(local, unsats))
      return FormulaCatalogue::bot;
    return h;
  };
};

Formula* FormulaSet::simplify(set<Formula*> &local, set<Formula*> &unsats)
{
  if (type == OR_SET) {
    Formula* f = getNegation();
    Formula* g = f->simplify(local, unsats);
    return (g == f ? this : g->getNegation());
  };

  // type == AND_SET

  if (isSubsumedBy(local, unsats))
    return FormulaCatalogue::top;

  if (getNegation()->isSubsumedBy(local, unsats))
    return FormulaCatalogue::bot;

  FormulaSet rs(AND_SET);
  set<Formula*> local2(local);
  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) {
    // simplifying *it

    Formula* f = *it;
    set<Formula*> local3(local2);
    set<Formula*>::iterator it2 = it;
    for (it2++; it2 != formulas.end(); it2++)
      local3.insert(*it2);

    Formula* fp = f->simplify(local3, unsats);
    if (fp == FormulaCatalogue::bot)
      return FormulaCatalogue::bot;
    if (fp == FormulaCatalogue::top)
      continue;

    rs.add(fp);
    local2.insert(fp);
  };

  if (!rs.getSetOfFormulas()->size())
    return FormulaCatalogue::top;

  return rs.normalize();
};

//------------------------------------------------------------------------------

Formula* Formula::simplifyRecursively(set<Formula*> &unsats)
{
  if (!Options::simplifyFormulas)
    return this;

  Formula* f = this;
  set<Formula*> local;

  Formula* old = f;
  f = f->simplify(local, unsats);
  if (f == old) return f;

  if (Options::simplifyFormulasTwice) {
    old = f;
    f = f->simplify(local, unsats);
    if (f == old) return f;
  }
  else return f;

  if (Options::simplifyFormulasRepeatedly) {
    int count = 3;
    while (true) {
      old = f;
      f = f->simplify(local, unsats);
      if (f == old) {
        #ifdef DEBUG
        cout << "Simplification did " << count << " rounds.\n\n";
        #endif
        return f;
      };
      assert(count++ < 1000); // just for debugging
    };
  };

  return f;
};

Formula* Formula::simplifyRecursively()
{
  set<Formula*> unsats;
  return simplifyRecursively(unsats);
};

//------------------------------------------------------------------------------

Formula* FormulaModalQuant::sat_simplify(set<Formula*> &unsats)
{
  if (Options::simplifyFormulas) {
    set<Formula*> local;
    return simplify(local, unsats);
  }
  else 
    return this;
};

Formula* FormulaSet::sat_simplify(set<Formula*> &unsats)
{
  if (!Options::literalElimination && !Options::simplifyFormulas)
    return this;

  Formula* f = this;

  if (Options::literalElimination) {
    map<Formula*,int> pos, neg, box, dia;
    countOccurrencesOfAtoms(pos, neg, box, dia);

    set<Formula*> lits;

    for (map<Formula*,int>::iterator it = pos.begin(); it != pos.end(); it++)
      if (neg.find(it->first) == neg.end())
        lits.insert(it->first);

    for (map<Formula*,int>::iterator it = neg.begin(); it != neg.end(); it++)
      if (pos.find(it->first) == pos.end())
        lits.insert(it->first->getNegation());

    if (lits.size())
      f = replaceAt1Level(lits);
  };

  set<Formula*> local;
  Formula* old = f;
  if (Options::simplifyFormulas)
    f = f->simplify(local, unsats);
  else return f;

  if (f == old) return f;

  old = f;
  if (Options::simplifyFormulasTwice) {
    f = f->sat_simplify(unsats);
    if (f == old) return f;
  }
  else return f;

  if (Options::simplifyFormulasRepeatedly) {
    int count = 3;
    while (true) {
      old = f;
      f = f->sat_simplify(unsats);
      if (f == old) {
        #ifdef DEBUG
        cout << "Simplification did " << count << " rounds.\n\n";
        #endif
        return f;
      };
      assert(count++ < 1000); // just for debugging
    };
  };

  return f;
};
