/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include <stdlib.h>
#include <math.h>
#include "tools.h"
#include "generator.h"

void AlcTBoxGen::writeParams(AlcTBoxGenType *gtype, FILE *pf)
{
  fputs(";;; This TBox is automatedly created using class AlcTBoxGen with the following parameters:\n", pf);

  fprintf(pf, ";;; randomSeed = %d\n", gtype->randomSeed);
  fprintf(pf, ";;; numOfRoles = %d\n", gtype->numOfRoles);
  fprintf(pf, ";;; numOfAtomicConcepts = %d\n", gtype->numOfAtomicConcepts);
  fprintf(pf, ";;; numOfConcepts = %d\n", gtype->numOfConcepts);

  fprintf(pf, ";;; probabilityAND  = %d (\%)\n", gtype->probabilityAND);
  fprintf(pf, ";;; probabilityOR   = %d (\%)\n", gtype->probabilityOR);
  fprintf(pf, ";;; probabilityALL  = %d (\%)\n", gtype->probabilityALL);
  fprintf(pf, ";;; probabilitySOME = %d (\%)\n", gtype->probabilitySOME);
  fprintf(pf, ";;; probabilityNOT  = %d (\%)\n", gtype->probabilityNOT);
  fprintf(pf, ";;; probabilityDUPL = %d (\%)\n", gtype->probabilityDUPL);

  fprintf(pf, ";;; firstArgIndexLowerBound = %f\n", gtype->firstArgIndexLowerBound);
  fprintf(pf, ";;; firstArgIndexUpperBound = %f\n", gtype->firstArgIndexUpperBound);
  fprintf(pf, ";;; secondArgIndexLowerBound = %f\n", gtype->secondArgIndexLowerBound);
  fprintf(pf, ";;; secondArgIndexUpperBound = %f\n", gtype->secondArgIndexUpperBound);
};

void AlcTBoxGen::generateTBox(AlcTBoxGenType *gtype, FILE *pf)
{
  srand(gtype->randomSeed);

  fputc('(', pf);

  for (int i = 1; i <= gtype->numOfRoles; i++)
    fprintf(pf, "(DEFINE-PRIMITIVE-ROLE R%d)\n", i);

  for (int i = 1; i <= gtype->numOfAtomicConcepts; i++) {
    fprintf(pf, "(DEFINE-PRIMITIVE-CONCEPT C%d)\n", 2*i-1);
    fprintf(pf, "(DEFINE-CONCEPT C%d (NOT C%d))\n", 2*i, 2*i-1);
  };

  int hundred = gtype->probabilityAND + gtype->probabilityOR +
                gtype->probabilityALL + gtype->probabilitySOME +
                gtype->probabilityNOT + gtype->probabilityDUPL;

  for (int i = 2 * gtype->numOfAtomicConcepts + 1; i <= gtype->numOfConcepts; i++) {
    fprintf(pf, "(DEFINE-CONCEPT C%d ", i);

    int p = rand() % hundred + 1;
    int numArgs;
    bool duplicate = false;
    if (p <= gtype->probabilityAND + gtype->probabilityOR) {
      numArgs = 2;
      fputs(p <= gtype->probabilityAND ? "(AND " : "(OR ", pf);
    }
    else {
      numArgs = 1;
      if (p <= gtype->probabilityAND + gtype->probabilityOR + 
               gtype->probabilityALL + gtype->probabilitySOME) {
        fputs(p <= (gtype->probabilityAND + gtype->probabilityOR + gtype->probabilityALL) ?
              "(ALL " : "(SOME ", pf);
        int k = rand() % gtype->numOfRoles + 1;
        fprintf(pf, "R%d ", k);
      }
      else if (p <= gtype->probabilityAND + gtype->probabilityOR +
                    gtype->probabilityALL + gtype->probabilitySOME +
                    gtype->probabilityNOT) {
        fputs("(NOT ", pf);
      }
      else
        duplicate = true;
    };

    if (duplicate)
      fprintf(pf, "C%d)\n", rand() % (i-1) + 1);
    else
      for (int j = 1; j <= numArgs; j++) {
        float lb = (j == 1 ? gtype->firstArgIndexLowerBound : gtype->secondArgIndexLowerBound);
        float ub = (j == 1 ? gtype->firstArgIndexUpperBound : gtype->secondArgIndexUpperBound);
        int lb2 = (int) floor(lb * (i-1));
        int ub2 = (int) floor(ub * (i-1));
        int dist = ub2 - lb2;
        if (!dist) 
          dist++;
        int k = rand() % dist + lb2 + 1;
        fprintf(pf, "C%d%s", k, (j == numArgs ? "))\n" : " "));
      }; // for
  }; // for

  fputc(')', pf);
};
