/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _NODE_H
#define _NODE_H

#include <set>
#include <map>
#include <queue>

#include "first.h"
#include "options.h"
#include "formula.h"

//------------------------------------------------------------------------------

#define labels_queue priority_queue<Formula*,vector<Formula*>,FormulaCompare>

typedef char NodeStatus;
typedef char NodeKind;
class NodeInfo {};

//------------------------------------------------------------------------------

class NodeInfo_CauseMaps4SB : public NodeInfo { // "cause"-maps for semantic branching
  public:
    map<Formula*,Formula*> byPosLit;
    map<Formula*,Formula*> byNegLit;
};

class NodeInfo_ParamsOfLE : public NodeInfo { // parameters for literals elimination
  public:
    set<Formula*> pureLiterals;
    set<Formula*> clsUnitFormulas; 
    set<Formula*> modUnitFormulas; 
};

class NodeInfo_CauseMap4LE : public NodeInfo { // "cause"-maps for literals elimination
  public:
    map<Formula*,Formula*> causeMap;
    set<Formula*> *unitFormulas;

    NodeInfo_CauseMap4LE()  { unitFormulas = NULL; };
    ~NodeInfo_CauseMap4LE() { if (unitFormulas) delete unitFormulas; };
};

class NodeInfo_4SR : public NodeInfo { // info for simplification rule
  public:
    Formula* simplifiedFormula;

    NodeInfo_4SR(Formula* _simplifiedFormula)
	{ simplifiedFormula = _simplifiedFormula; };
};

//------------------------------------------------------------------------------

class Node {
  friend class Graph;
  friend class NodeCompare;

  protected:
    int id;
    Formula* content;
    NodeStatus status;
    NodeKind kind;
    char rule; // identification of the applied rule(s)

    bool reachable; // heuristically reachabe from "tau"
    unsigned short depth; // depth of the node in the skeleton tree
    unsigned short numEssentialChildren; // #children with status different than SAT and UNSAT

    Formula* principal; // principal == NULL => AND_NODE
    Node* mainParent;
    set<Node*> parents; 
    map<Node*,Formula*> edges; // edge = (child,label)

    double priority;

    NodeInfo *otherInfo;

  public:
    Node();

    NodeStatus getStatus() { return status; };
    Formula* getContent() { return content; };
    char getRule() { return rule; };

    virtual bool chooseRule(labels_queue &labels) = 0; // updates rule, principal, labels, otherInfo
    virtual Formula* applyRule(Formula* label, Formula* generalAssumptions = NULL) = 0;
    virtual Formula* computeUnsatCore(Formula* generalAssumptions, Node* child = NULL)
		{ /* if we want to ignore: */ return content; };
};

#endif
