/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include "first.h"
#include "tools.h"
#include "options.h"

bool Options::usingOptionsFile;

bool Options::hsExpanding;
bool Options::dfsExpanding;
bool Options::bfsExpanding;

bool Options::useUnsatCores;
double Options::unsatCoreThreshold4SubsetChecking;

bool Options::subsetChecking;
bool Options::subsetChecking1;
bool Options::subsetChecking2;
int Options::maxLenLimitForKeptUnsat;
bool Options::localSupersetChecking;

bool Options::cutoff;
bool Options::checkSubsumption4NormalForm;
bool Options::checkSubsumption4FirstLayer;
int  Options::maxSetSize4CheckingSubsumption;

bool Options::literalElimination;
int  Options::literalEliminationThreshold;
bool Options::semanticBranching1;
int  Options::semanticBranching1Threshold;
bool Options::semanticBranching2;
int  Options::semanticBranching2Threshold;

bool Options::usingSize4OrBranching;
int  Options::weightingMethod;
double Options::weightOfAll;
double Options::weightOfSome;
double Options::weightOfAnd;
double Options::weightOfOr;
double Options::weightOfMaxElem;
double Options::expansionFactorOfNumOfWorlds;

bool Options::simplifyFormulas;
bool Options::simplifyFormulasTwice;
bool Options::simplifyFormulasRepeatedly;

bool Options::simplifyNodes;
double Options::relativeNodeSize4Simp;

int Options::minExpansionsBDSN; 
double Options::minExpansionsRateBDSN;

bool Options::randomDFS;
bool Options::usingSize4SelectingDfsBranch;
int Options::dfsExpandingQueueSizeThreshold;
double Options::dfsExpandingQueueSizeRatio;

double Options::priorityOfPropagationDegree;
double Options::propagationDegreeGranule;
double Options::satProbability;
double Options::weightOfDynamicSatProbability;

int Options::numOfGroupsOfFormulas;
double Options::priorityOfContentSize;
double Options::priorityOfContentWeight;

double Options::priorityOfDFS;
double Options::priorityOfNodeDepth;

int Options::minExpansionsBREQ;
double Options::minExpansionsRateBREQ;
double Options::minDeletedBackEdgesRateBREQ;

int Options::minFormulasBeforeCompactingCatalogue;
double Options::expectedCatalogueCompactingGain;
double Options::catalogueExpandingRateLimit;

double Options::formulaCatalogueMaxFillRatio;
double Options::hashPreallocateFactor;

bool Options::breakWhenUnsuccessful;

int Options::secondsBeforeSIGKILL;
bool Options::verbose;

//-----------------------------------------------------------------------------

string Options::optionsFileName = "TGC-options";

void Options::init()
{
  usingOptionsFile = false;

  hsExpanding  = false;
  dfsExpanding = true;
  bfsExpanding = false;

  useUnsatCores = true;
  unsatCoreThreshold4SubsetChecking = 0.9999;

  subsetChecking = true;
  subsetChecking1 = true;
  subsetChecking2 = true;
  maxLenLimitForKeptUnsat = 5;
  localSupersetChecking = true;

  cutoff = true;
  checkSubsumption4NormalForm = false;
  checkSubsumption4FirstLayer = false;
  maxSetSize4CheckingSubsumption = 3;

  literalElimination = true;
  literalEliminationThreshold = 1;
  semanticBranching1 = true;
  semanticBranching1Threshold = 2;
  semanticBranching2 = true;
  semanticBranching2Threshold = 3;

  usingSize4OrBranching = false;
  weightingMethod = WM_SUMMING;
  weightOfAll  = 1.02;
  weightOfSome = 1.05;
  weightOfAnd  = 1.0;
  weightOfOr   = 4.0;
  weightOfMaxElem = 0.5;
  expansionFactorOfNumOfWorlds = 1.1;

  simplifyFormulas = false;
  simplifyFormulasTwice = false;
  simplifyFormulasRepeatedly = false;

  simplifyNodes = false;
  relativeNodeSize4Simp = 0.01;

  minExpansionsBDSN = 1000;
  minExpansionsRateBDSN = 0.3;

  randomDFS = false;
  usingSize4SelectingDfsBranch = false;
  dfsExpandingQueueSizeThreshold = 1000;
  dfsExpandingQueueSizeRatio = 2.0;

  priorityOfPropagationDegree = 1;
  propagationDegreeGranule = 0.01;
  satProbability = 0.5;
  weightOfDynamicSatProbability = 0;

//numOfGroupsOfFormulas = 100;
  numOfGroupsOfFormulas = 1000;
  priorityOfContentSize = 0;
//priorityOfContentSize = -0.01;
//priorityOfContentWeight = 0;
  priorityOfContentWeight = -0.01;

  priorityOfDFS = 0.00001;
  priorityOfNodeDepth = -0.00000;
//priorityOfNodeDepth = -0.00001;

  minExpansionsBREQ = 300;
  minExpansionsRateBREQ = 0.5;
  minDeletedBackEdgesRateBREQ = 0.3;

  minFormulasBeforeCompactingCatalogue = 10000;
  expectedCatalogueCompactingGain = 0.33;
  catalogueExpandingRateLimit = 3.0;

  formulaCatalogueMaxFillRatio = 1.0;
  hashPreallocateFactor = 12.0;

  breakWhenUnsuccessful = false;

  secondsBeforeSIGKILL = 3;
  verbose = true;

  if (optionsFileName.compare("") != 0) {
    FILE *pf;
    char msg[256];
    if (!(pf = fopen(optionsFileName.c_str(), "r"))) {
      sprintf(msg, "Error on opening the option file \"%s\".", 
              optionsFileName.c_str());
      MSG(msg);
      MSG("Using default options.");
    }
    else {
      usingOptionsFile = true;
      readAllOptions(pf);
      fclose(pf);
    };
  };

  if (dfsExpanding)
    if (hsExpanding || bfsExpanding) {
      hsExpanding = false;
      bfsExpanding = false;
      MSG("Chosen DFS.");
    };

  if (hsExpanding)
    if (bfsExpanding) {
      bfsExpanding = false;
      MSG("Chosen Heuristic Search.");
    };

  if (!priorityOfPropagationDegree) {
    propagationDegreeGranule = 1;
    satProbability = 0.5;
    weightOfDynamicSatProbability = 0;
  };

  if (priorityOfDFS > 0 && priorityOfContentSize != 0)
    priorityOfContentWeight = 0;

  if (weightOfMaxElem < 0 || weightOfMaxElem > 1.0) {
    MSG("Given a wrong value for weightOfMaxElem.\nUsing the default value for it.");
    weightOfMaxElem = 0.5;
  };
};

//-----------------------------------------------------------------------------

void Options::readAllOptions(FILE *pf)
{
  while (!feof(pf))
    if (!readOption(pf))
      return;
};

#define _TYPE_BOOL	1
#define _TYPE_INT	2
#define _TYPE_DOUBLE	3

#define try_reading_option(opt, name, type) { \
  if (s.compare(name) == 0) { \
    s = getSym(pf); \
    if (s.compare("=") != 0) { \
      char msg[256]; \
      sprintf(msg, "Expected \"=\" after \"%s\".", name); \
      MSG(msg); \
      return false; \
    }; \
    \
    s = getSym(pf); \
    if (type == _TYPE_BOOL) { \
      if (s.compare("true") != 0 && s.compare("TRUE") != 0 && \
          s.compare("false") != 0 && s.compare("FALSE") != 0) { \
        char msg[256]; \
        sprintf(msg, "Expected \"true\" or \"false\" after \"%s = \".", name); \
        MSG(msg); \
        return false; \
      }; \
      opt = (s.compare("true") == 0 || s.compare("TRUE") == 0); \
    } \
    else if (type == _TYPE_INT) { \
      opt = atoi(s.c_str()); \
    } \
    else if (type == _TYPE_DOUBLE) { \
      opt = atof(s.c_str()); \
    } \
    else assert(false); \
    \
    return true; \
  }; \
}

bool Options::readOption(FILE *pf)
{
  string s = getSym(pf);

  if (s.compare("EOF") == 0)
    return true;

  if (s.compare("ERROR") == 0) {
    MSG("Error on trying to read an option name.");
    return false;
  };

  try_reading_option(hsExpanding, "hsExpanding", _TYPE_BOOL);
  try_reading_option(dfsExpanding, "dfsExpanding", _TYPE_BOOL);
  try_reading_option(bfsExpanding, "bfsExpanding", _TYPE_BOOL);

  try_reading_option(useUnsatCores, "useUnsatCores", _TYPE_BOOL);
  try_reading_option(unsatCoreThreshold4SubsetChecking, 
                     "unsatCoreThreshold4SubsetChecking", _TYPE_DOUBLE);

  try_reading_option(subsetChecking, "subsetChecking", _TYPE_BOOL);
  try_reading_option(subsetChecking1, "subsetChecking1", _TYPE_BOOL);
  try_reading_option(subsetChecking2, "subsetChecking2", _TYPE_BOOL);
  try_reading_option(maxLenLimitForKeptUnsat, "maxLenLimitForKeptUnsat", _TYPE_INT);
  try_reading_option(localSupersetChecking, "localSupersetChecking", _TYPE_BOOL);

  try_reading_option(cutoff, "cutoff", _TYPE_BOOL);
  try_reading_option(checkSubsumption4NormalForm, "checkSubsumption4NormalForm", _TYPE_BOOL);
  try_reading_option(checkSubsumption4FirstLayer, "checkSubsumption4FirstLayer", _TYPE_BOOL);
  try_reading_option(maxSetSize4CheckingSubsumption, "maxSetSize4CheckingSubsumption", _TYPE_INT);

  try_reading_option(literalElimination, "literalElimination", _TYPE_BOOL);
  try_reading_option(literalEliminationThreshold, "literalEliminationThreshold", _TYPE_INT);
  try_reading_option(semanticBranching1, "semanticBranching1", _TYPE_BOOL);
  try_reading_option(semanticBranching1Threshold, "semanticBranching1Threshold", _TYPE_INT);
  try_reading_option(semanticBranching2, "semanticBranching2", _TYPE_BOOL);
  try_reading_option(semanticBranching2Threshold, "semanticBranching2Threshold", _TYPE_INT);

  try_reading_option(usingSize4OrBranching, "usingSize4OrBranching", _TYPE_BOOL);
  try_reading_option(weightingMethod, "weightingMethod", _TYPE_INT);
  try_reading_option(weightOfAll, "weightOfAll", _TYPE_DOUBLE);
  try_reading_option(weightOfSome, "weightOfSome", _TYPE_DOUBLE);
  try_reading_option(weightOfAnd, "weightOfAnd", _TYPE_DOUBLE);
  try_reading_option(weightOfOr, "weightOfOr", _TYPE_DOUBLE);
  try_reading_option(weightOfMaxElem, "weightOfMaxElem", _TYPE_DOUBLE);
  try_reading_option(expansionFactorOfNumOfWorlds, "expansionFactorOfNumOfWorlds", _TYPE_DOUBLE);

  try_reading_option(simplifyFormulas, "simplifyFormulas", _TYPE_BOOL);
  try_reading_option(simplifyFormulasTwice, "simplifyFormulasTwice", _TYPE_BOOL);
  try_reading_option(simplifyFormulasRepeatedly, "simplifyFormulasRepeatedly", _TYPE_BOOL);

  try_reading_option(simplifyNodes, "simplifyNodes", _TYPE_BOOL);
  try_reading_option(relativeNodeSize4Simp, "relativeNodeSize4Simp", _TYPE_DOUBLE);

  try_reading_option(minExpansionsBDSN, "minExpansionsBDSN", _TYPE_INT);
  try_reading_option(minExpansionsRateBDSN, "minExpansionsRateBDSN", _TYPE_DOUBLE);

  try_reading_option(randomDFS, "randomDFS", _TYPE_BOOL);
  try_reading_option(usingSize4SelectingDfsBranch, "usingSize4SelectingDfsBranch", _TYPE_BOOL);
  try_reading_option(dfsExpandingQueueSizeThreshold, "dfsExpandingQueueSizeThreshold", _TYPE_INT);
  try_reading_option(dfsExpandingQueueSizeRatio, "dfsExpandingQueueSizeRatio", _TYPE_DOUBLE);

  try_reading_option(priorityOfPropagationDegree, "priorityOfPropagationDegree", _TYPE_DOUBLE);
  try_reading_option(propagationDegreeGranule, "propagationDegreeGranule", _TYPE_DOUBLE);
  try_reading_option(satProbability, "satProbability", _TYPE_DOUBLE);
  try_reading_option(weightOfDynamicSatProbability, "weightOfDynamicSatProbability", _TYPE_DOUBLE);

  try_reading_option(numOfGroupsOfFormulas, "numOfGroupsOfFormulas", _TYPE_INT);
  try_reading_option(priorityOfContentSize, "priorityOfContentSize", _TYPE_DOUBLE);
  try_reading_option(priorityOfContentWeight, "priorityOfContentWeight", _TYPE_DOUBLE);

  try_reading_option(priorityOfDFS, "priorityOfDFS", _TYPE_DOUBLE);
  try_reading_option(priorityOfNodeDepth, "priorityOfNodeDepth", _TYPE_DOUBLE);

  try_reading_option(minExpansionsBREQ, "minExpansionsBREQ", _TYPE_INT);
  try_reading_option(minExpansionsRateBREQ, "minExpansionsRateBREQ", _TYPE_DOUBLE);
  try_reading_option(minDeletedBackEdgesRateBREQ, "minDeletedBackEdgesRateBREQ", _TYPE_DOUBLE);

  try_reading_option(minFormulasBeforeCompactingCatalogue, 
                     "minFormulasBeforeCompactingCatalogue", _TYPE_INT);
  try_reading_option(expectedCatalogueCompactingGain, 
                     "expectedCatalogueCompactingGain", _TYPE_DOUBLE);
  try_reading_option(catalogueExpandingRateLimit, "catalogueExpandingRateLimit", _TYPE_DOUBLE);

  try_reading_option(formulaCatalogueMaxFillRatio, "formulaCatalogueMaxFillRatio", _TYPE_DOUBLE);
  try_reading_option(hashPreallocateFactor, "hashPreallocateFactor", _TYPE_DOUBLE);

  try_reading_option(breakWhenUnsuccessful, "breakWhenUnsuccessful", _TYPE_BOOL);

  try_reading_option(secondsBeforeSIGKILL, "secondsBeforeSIGKILL", _TYPE_INT);
  try_reading_option(verbose, "verbose", _TYPE_BOOL);

  char msg[256];
  sprintf(msg, "Encountered an unexpected symbol \"%s\".\nIngoring the rest of the option file.", s.c_str());
  MSG(msg);
  return false;
};

//-----------------------------------------------------------------------------

bool Options::readAlcGenParams(AlcTBoxGenType *gtype, const char* fileName)
{
  FILE *pf;
  char msg[256];
  if (!(pf = fopen(fileName, "r"))) {
    sprintf(msg, "Error on opening the parameter file \"%s\".", fileName);
    MSG(msg);
    return false;
  };

  while (!feof(pf))
    if (!readAlcGenParam(gtype, pf)) {
      fclose(pf);
      return false;
    };

  fclose(pf);
  return true;
};

bool Options::readAlcGenParam(AlcTBoxGenType *gtype, FILE *pf)
{
  string s = getSym(pf);

  if (s.compare("EOF") == 0)
    return true;

  if (s.compare("ERROR") == 0) {
    MSG("Error on trying to read a parameter name.");
    return false;
  };

  try_reading_option(gtype->randomSeed, "randomSeed", _TYPE_INT);
  try_reading_option(gtype->numOfRoles, "numOfRoles", _TYPE_INT);
  try_reading_option(gtype->numOfAtomicConcepts, "numOfAtomicConcepts", _TYPE_INT);
  try_reading_option(gtype->numOfConcepts, "numOfConcepts", _TYPE_INT);

  try_reading_option(gtype->probabilityAND, "probabilityAND", _TYPE_INT);
  try_reading_option(gtype->probabilityOR, "probabilityOR", _TYPE_INT);
  try_reading_option(gtype->probabilityALL, "probabilityALL", _TYPE_INT);
  try_reading_option(gtype->probabilitySOME, "probabilitySOME", _TYPE_INT);
  try_reading_option(gtype->probabilityNOT, "probabilityNOT", _TYPE_INT);
  try_reading_option(gtype->probabilityDUPL, "probabilityDUPL", _TYPE_INT);

  try_reading_option(gtype->firstArgIndexLowerBound, "firstArgIndexLowerBound", _TYPE_DOUBLE);
  try_reading_option(gtype->firstArgIndexUpperBound, "firstArgIndexUpperBound", _TYPE_DOUBLE);
  try_reading_option(gtype->secondArgIndexLowerBound, "secondArgIndexLowerBound", _TYPE_DOUBLE);
  try_reading_option(gtype->secondArgIndexUpperBound, "secondArgIndexUpperBound", _TYPE_DOUBLE);

  char msg[256];
  sprintf(msg, "Encountered an unexpected symbol \"%s\".\nIngoring the rest of the parameter file.", s.c_str());
  MSG(msg);
  return false;
}

//-----------------------------------------------------------------------------

void Options::omitSeperators(FILE *pf)
{
  while (!feof(pf)) {
    char ch = fgetc(pf);
    if (ch != ' ' && ch != '\t' && ch != '\n' && ch != ';' && ch != '%') {
      ungetc(ch, pf);
      return;
    }
    else if (ch == '%')
      while (!feof(pf) && ch != '\n')
        ch = fgetc(pf);
  };
};

string Options::getSym(FILE *pf)
{
  string s;

  omitSeperators(pf);
  char ch = fgetc(pf);

  switch (ch) {
    case EOF : return (s = "EOF");
    case '=' : return (s = "=");
  };

  if (!isalnum(ch) && ch != '-')
    return (s = "ERROR");
  s.append(1,ch);

  while (!feof(pf)) {
    ch = fgetc(pf);
    if (!isalnum(ch) && ch != '.') {
      ungetc(ch,pf);
      return s;
    };
    s.append(1,ch);
  };

  return s;
};
