/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _OPTIONS_H
#define _OPTIONS_H

#include <string>
#include <stdio.h>
#include "generator.h"

// weighting methods:
#define WM_SUMMING		1
#define WM_DIAMOND_DEPTH	2
#define WM_NO_OF_WORLDS		3


class Options {
  public:

    static bool usingOptionsFile;

  // general options for effectiveness:

    static bool hsExpanding;
    static bool dfsExpanding;
    static bool bfsExpanding;

    static bool useUnsatCores;
    static double unsatCoreThreshold4SubsetChecking;

    static bool subsetChecking;
    static bool subsetChecking1;
    static bool subsetChecking2;
    static int  maxLenLimitForKeptUnsat;
    static bool localSupersetChecking;

    static bool cutoff;
    static bool checkSubsumption4NormalForm;
    static bool checkSubsumption4FirstLayer;
    static int  maxSetSize4CheckingSubsumption;

    static bool literalElimination;
    static int  literalEliminationThreshold;
    static bool semanticBranching1;
    static int  semanticBranching1Threshold;
    static bool semanticBranching2;
    static int  semanticBranching2Threshold;

    static bool usingSize4OrBranching; // if not, use "weight"
    static int  weightingMethod; // WM_SUMMING, WM_DIAMOND_DEPTH, or WM_NO_OF_WORLDS
    // for the case weightingMethod == WM_SUMMING
    static double weightOfAll;
    static double weightOfSome;
    static double weightOfAnd;
    static double weightOfOr;
    // for the case weightingMethod != WM_SUMMING
    static double weightOfMaxElem;
    // for the case weightingMethod == WM_NO_OF_WORLDS
    static double expansionFactorOfNumOfWorlds;

    static bool simplifyFormulas;
    static bool simplifyFormulasTwice;
    static bool simplifyFormulasRepeatedly;

    static bool simplifyNodes;
    static double relativeNodeSize4Simp;

    // BDSN: Before Detecting SAT Nodes
    static int minExpansionsBDSN; 
    static double minExpansionsRateBDSN;

  // DFS parameters:

    static bool randomDFS;
    static bool usingSize4SelectingDfsBranch; // if not, use "weight"
    static int dfsExpandingQueueSizeThreshold;
    static double dfsExpandingQueueSizeRatio;

  // HS parameters:

    static double priorityOfPropagationDegree; 
    static double propagationDegreeGranule;
    static double satProbability;
    static double weightOfDynamicSatProbability;

    static int numOfGroupsOfFormulas;
    static double priorityOfContentSize;
    static double priorityOfContentWeight;

    static double priorityOfDFS;
    static double priorityOfNodeDepth;

    // BREQ: Before Recomputing Expanding Queue
    static int minExpansionsBREQ; 
    static double minExpansionsRateBREQ;
    static double minDeletedBackEdgesRateBREQ;

  // for maintaining FormulaCatalogue:

    static int minFormulasBeforeCompactingCatalogue;
    static double expectedCatalogueCompactingGain;
    static double catalogueExpandingRateLimit;

    static double formulaCatalogueMaxFillRatio;
    static double hashPreallocateFactor;

  // for dealing with batch input files
    static bool breakWhenUnsuccessful;

  // others
    static int secondsBeforeSIGKILL;
    static bool verbose;

  public:
    static string optionsFileName;
    static void init();
    static bool readAlcGenParams(AlcTBoxGenType *gtype, const char* fileName);

  private:
    static string getSym(FILE *pf);
    static void omitSeperators(FILE *pf);
    static bool readOption(FILE *pf);
    static void readAllOptions(FILE *pf);
    static bool readAlcGenParam(AlcTBoxGenType *gtype, FILE *pf);
};

#endif
