/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include "statistics.h"

ChildInfo* _stat;

Statistics::Statistics()
{
  numOfSatResults = 0;
  numOfUnsatResults = 0;
  numOfUnknownResults = 0;

  numOfExpansions = 0;

  totalTime = 0;

  applyingRules = 0;
  choosingRules = 0;

  computingUnsatCores = 0;
  usingUnsatCores = 0;

  subsetChecking = 0;
  subsetChecking1 = 0;
  subsetChecking2 = 0;
  updatingUnsats = 0;
  localSupersetChecking = 0;

  eliminatingLiterals = 0;
  semanticBranching1 = 0;
  semanticBranching2 = 0;

  detectingSatNodes = 0;

  recomputingExpandingQueue = 0;
  computingNodePriorities = 0;
  updatingReachabilityRelation = 0;

  mergingNodes = 0;
  compactingFormulaCatalogue = 0;

  simplifyingFormulas = 0;
};

void Statistics::add(ChildInfo* lstat, bool ignoreUNKNOWN)
{
  numOfSatResults += (lstat->result == SAT ? 1 : 0);
  numOfUnsatResults += (lstat->result == UNSAT ? 1 : 0);
  numOfUnknownResults += (lstat->result == UNKNOWN ? 1 : 0);

  if (ignoreUNKNOWN && lstat->result == UNKNOWN) return;

  numOfExpansions += lstat->numOfExpansions;

  totalTime += ((double) lstat->totalTime[0]) / CLOCKS_PER_SEC;

  applyingRules += ((double) lstat->applyingRules[0]) / CLOCKS_PER_SEC;
  choosingRules += ((double) lstat->choosingRules[0]) / CLOCKS_PER_SEC;

  computingUnsatCores += ((double) lstat->computingUnsatCores[0]) / CLOCKS_PER_SEC;
  usingUnsatCores += ((double) lstat->usingUnsatCores[0]) / CLOCKS_PER_SEC;

  subsetChecking += ((double) lstat->subsetChecking[0]) / CLOCKS_PER_SEC;
  subsetChecking1 += ((double) lstat->subsetChecking1[0]) / CLOCKS_PER_SEC;
  subsetChecking2 += ((double) lstat->subsetChecking2[0]) / CLOCKS_PER_SEC;
  updatingUnsats += ((double) lstat->updatingUnsats[0]) / CLOCKS_PER_SEC;
  localSupersetChecking += ((double) lstat->localSupersetChecking[0]) / CLOCKS_PER_SEC;

  eliminatingLiterals += ((double) lstat->eliminatingLiterals[0]) / CLOCKS_PER_SEC;
  semanticBranching1 += ((double) lstat->semanticBranching1[0]) / CLOCKS_PER_SEC;
  semanticBranching2 += ((double) lstat->semanticBranching2[0]) / CLOCKS_PER_SEC;

  detectingSatNodes += ((double) lstat->detectingSatNodes[0]) / CLOCKS_PER_SEC;

  recomputingExpandingQueue += ((double) lstat->recomputingExpandingQueue[0]) / CLOCKS_PER_SEC;
  computingNodePriorities += ((double) lstat->computingNodePriorities[0]) / CLOCKS_PER_SEC;
  updatingReachabilityRelation += 
    ((double) lstat->updatingReachabilityRelation[0]) / CLOCKS_PER_SEC;

  mergingNodes += ((double) lstat->mergingNodes[0]) / CLOCKS_PER_SEC;
  compactingFormulaCatalogue += ((double) lstat->compactingFormulaCatalogue[0]) / CLOCKS_PER_SEC;

  simplifyingFormulas += ((double) lstat->simplifyingFormulas[0]) / CLOCKS_PER_SEC;
};
