/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _STATISTICS_H
#define _STATISTICS_H

#include <ctime>
#include "first.h"

typedef struct {
    bool exited;
    char result; // SAT, UNSAT, or UNKNOWN

    int numOfExpansions; 

    clock_t totalTime[2];

    clock_t applyingRules[2];
    clock_t choosingRules[2];

    clock_t computingUnsatCores[2];
    clock_t usingUnsatCores[2];

    clock_t subsetChecking[2];
    clock_t subsetChecking1[2];
    clock_t subsetChecking2[2];
    clock_t localSupersetChecking[2];
    clock_t updatingUnsats[2];

    clock_t eliminatingLiterals[2];
    clock_t semanticBranching1[2];
    clock_t semanticBranching2[2];

    clock_t detectingSatNodes[2];

    clock_t recomputingExpandingQueue[2];
    clock_t computingNodePriorities[2];
    clock_t updatingReachabilityRelation[2];

    clock_t mergingNodes[2];
    clock_t compactingFormulaCatalogue[2];

    clock_t simplifyingFormulas[2];

} ChildInfo;

extern ChildInfo* _stat;

#define STATISTICS_INIT() { \
  _stat->exited = false; \
  _stat->result = UNKNOWN; \
  _stat->numOfExpansions = 0; \
  \
  _stat->totalTime[0] = 0; \
  \
  _stat->applyingRules[0] = 0; \
  _stat->choosingRules[0] = 0; \
  \
  _stat->computingUnsatCores[0] = 0; \
  _stat->usingUnsatCores[0] = 0; \
  \
  _stat->subsetChecking[0] = 0; \
  _stat->subsetChecking1[0] = 0; \
  _stat->subsetChecking2[0] = 0; \
  _stat->localSupersetChecking[0] = 0; \
  _stat->updatingUnsats[0] = 0; \
  \
  _stat->eliminatingLiterals[0] = 0; \
  _stat->semanticBranching1[0] = 0; \
  _stat->semanticBranching2[0] = 0; \
  \
  _stat->detectingSatNodes[0] = 0; \
  \
  _stat->recomputingExpandingQueue[0] = 0; \
  _stat->computingNodePriorities[0] = 0; \
  _stat->updatingReachabilityRelation[0] = 0; \
  \
  _stat->mergingNodes[0] = 0; \
  _stat->compactingFormulaCatalogue[0] = 0; \
  \
  _stat->simplifyingFormulas[0] = 0; \
  \
  _stat->totalTime[1] = 0; \
  \
  _stat->applyingRules[1] = 0; \
  _stat->choosingRules[1] = 0; \
  \
  _stat->computingUnsatCores[1] = 0; \
  _stat->usingUnsatCores[1] = 0; \
  \
  _stat->subsetChecking[1] = 0; \
  _stat->subsetChecking1[1] = 0; \
  _stat->subsetChecking2[1] = 0; \
  _stat->localSupersetChecking[1] = 0; \
  _stat->updatingUnsats[1] = 0; \
  \
  _stat->eliminatingLiterals[1] = 0; \
  _stat->semanticBranching1[1] = 0; \
  _stat->semanticBranching2[1] = 0; \
  \
  _stat->detectingSatNodes[1] = 0; \
  \
  _stat->recomputingExpandingQueue[1] = 0; \
  _stat->computingNodePriorities[1] = 0; \
  _stat->updatingReachabilityRelation[1] = 0; \
  \
  _stat->mergingNodes[1] = 0; \
  _stat->compactingFormulaCatalogue[1] = 0; \
  \
  _stat->simplifyingFormulas[1] = 0; \
}

#define STATISTICS_SET(attr, value) { _stat->attr = value; }
#define STATISTICS_INC(attr) { _stat->attr++; }
#define STATISTICS_INC2(attr, value) { _stat->attr[0] += value; }

#define STATISTICS_START(attr) { \
	assert(!_stat->attr[1]); \
	_stat->attr[1] = clock() + 1; \
}

#define STATISTICS_STOP(attr) { \
	_stat->attr[0] += clock() + 1 - _stat->attr[1]; \
	_stat->attr[1] = 0; \
}


class Statistics {
  public: 
    int numOfSatResults;
    int numOfUnsatResults;
    int numOfUnknownResults;

    int numOfExpansions;

    double totalTime;

    double applyingRules;
    double choosingRules;

    double computingUnsatCores;
    double usingUnsatCores;

    double subsetChecking;
    double subsetChecking1;
    double subsetChecking2;
    double localSupersetChecking;
    double updatingUnsats;

    double eliminatingLiterals;
    double semanticBranching1;
    double semanticBranching2;

    double detectingSatNodes;

    double recomputingExpandingQueue;
    double computingNodePriorities;
    double updatingReachabilityRelation;

    double mergingNodes;
    double compactingFormulaCatalogue;

    double simplifyingFormulas;

  public:
    Statistics();
    void add(ChildInfo* lstat, bool ignoreUNKNOWN = false);
};

#endif
