/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdint.h>     /* defines uint32_t etc */
#include <set>

using namespace std;

#include "tools.h"

/* I acknowledge the use of parts of lookup3.c written by Bob Jenkins
   for defining the following hashword functions */

#define rot(x,k) (((x)<<(k)) | ((x)>>(32-(k))))

#define mix(a,b,c) \
{ \
  a -= c;  a ^= rot(c, 4);  c += b; \
  b -= a;  b ^= rot(a, 6);  a += c; \
  c -= b;  c ^= rot(b, 8);  b += a; \
  a -= c;  a ^= rot(c,16);  c += b; \
  b -= a;  b ^= rot(a,19);  a += c; \
  c -= b;  c ^= rot(b, 4);  b += a; \
}

#define final(a,b,c) \
{ \
  c ^= b; c -= rot(b,14); \
  a ^= c; a -= rot(c,11); \
  b ^= a; b -= rot(a,25); \
  c ^= b; c -= rot(b,16); \
  a ^= c; a -= rot(c,4);  \
  b ^= a; b -= rot(a,14); \
  c ^= b; c -= rot(b,24); \
}

#define HASHWORD_INITVAL	0

uint32_t hashword(int type)
{
  uint32_t a,b,c;
  uint32_t length = 1;
  a = b = c = 0xdeadbeef + (length<<2) + HASHWORD_INITVAL;

  a += (uint32_t) type;

  final(a,b,c);
  return c;
}

uint32_t hashword(int type, uint32_t x)
{
  uint32_t a,b,c;
  uint32_t length = 2;
  a = b = c = 0xdeadbeef + (length<<2) + HASHWORD_INITVAL;

  b += x;
  a += (uint32_t) type;

  final(a,b,c);
  return c;
}

uint32_t hashword(int type, uint32_t x, uint32_t y)
{
  uint32_t a,b,c;
  uint32_t length = 3;
  a = b = c = 0xdeadbeef + (length<<2) + HASHWORD_INITVAL;

  c += y;
  b += x;
  a += (uint32_t) type;

  final(a,b,c);
  return c;
}

uint32_t hashword(int type, set<Formula*> &s)
{
  uint32_t a,b,c;
  uint32_t length = s.size();
  a = b = c = 0xdeadbeef + ((length+1)<<2) + HASHWORD_INITVAL;

  set<Formula*>::iterator it = s.begin();

  while (length >= 3)
  {
    a += uint32_t (*(it++));
    b += uint32_t (*(it++));
    c += uint32_t (*(it++));
    mix(a,b,c);
    length -= 3;
  }

  switch(length) {
    case 0 : 
      a += (uint32_t) type; 
      final(a,b,c);
      return c;

    case 1 : 
      a += uint32_t (*(it++));
      b += (uint32_t) type; 
      final(a,b,c);
      return c;

    case 2 : 
      a += uint32_t (*(it++));
      b += uint32_t (*(it++));
      c += (uint32_t) type; 
      final(a,b,c);
      return c;
  };
};

// ----------------------------------------------------------------------------
// floor_log2 and ceil_log2 were provided by Lukasz Czajka

unsigned floor_log2(unsigned k)
{
  assert(k != 0);
  unsigned r = 0;
  k = k >> 1;
  while (k != 0)
  {
    ++r;
    k = k >> 1;
  }
  return r;
}

unsigned ceil_log2(unsigned k)
{
  assert(k != 0);
  int r = 0;
  int i = -1;
  if (k & 1 != 0)
  {
    i = 0;
  }
  k = k >> 1;
  while (k != 0)
  {
    ++r;
    if (k & 1 != 0 && i == -1)
    {
      i = r;
    }
    k = k >> 1;
  }
  assert (i >= 0);
  assert (i <= r);
  return r + (i == r ? 0 : 1);
}

